from importlib.machinery import EXTENSION_SUFFIXES as _EXTENSION_SUFFIXES
from pathlib import Path, PurePath

__all__ = [
    "Package",
    "Project",
    "PydFile",
    "LiteralXML",
    "ConditionalValue",
    "Property",
    "ItemDefinition",
    "PyFile",
    "SourceFile",
    "CSourceFile",
    "IncludeFile",
    "File",
    "RemoveFile",
]


class _Project:
    r"""Base class of compilable projects. Do not use directly."""
    options = {}

    def __init__(self, name, *members, project_file=None, source="", **options):
        self.name = name
        self.source = source
        self.options = {**self.options, **options}
        self.project_file = project_file
        self.members = list(members)

    def __iter__(self):
        for m in self.members:
            yield m
            try:
                it = iter(m)
            except TypeError:
                pass
            else:
                yield from it


class Package(_Project):
    r"""Represents a Python package.

Each Package represents a directory in the final distribution. Its
members may be other packages, files, or configuration settings.

Packages and files that are members of a Package will be installed
to the package directory. Note that only explicitly listed contents
will be included.

Specify `project_file` to override all generation with a fixed
MSBuild project file. This also overrides additional options and
sdist generation.

Specify `source` to find sources in a subdirectory without impacting
the install structure.

Other options will be added to the project as properties.
"""


class PydFile(_Project):
    r"""Represents a .pyd module.

Each PydFile represents a single .pyd file in the final distribution.
It must be included in a Package, and the built module will be
included in that package's directory.

Its members should be source files or projects. Directly referenced
files will be included in sdists, but transitive references will not.

Specify `project_file` to override all generation with a fixed
MSBuild project file. This also overrides additional options and
sdist generation.

Specify `source` to find sources in a subdirectory.

Other options will be added to the project as properties.
"""
    options = {"TargetExt": None}


class LiteralXML:
    r"""Literal string to insert into generated project file.

This must be valid XML and may use any MSBuild syntax. It will be
inserted under the top-level <Project> element.
"""
    members = ()

    def __init__(self, xml_string):
        self.name = "##xml"
        self.xml = xml_string


class ConditionalValue:
    r"""Add a condition to any property value.

This may be applied to any value that will be written into a project
file.

Specify `condition` to directly control the test, or `if_empty` to
check for an existing value.

Specify either `prepend` or `append` (or both, if you have a reason)
to concatenate the existing property value. Remember to include the
appropriate separator character in your value.
"""
    has_condition = True

    def __init__(self, value, *, condition=None, if_empty=False, prepend=False, append=False):
        self.value = value
        # Ensure condition has a True value so we can test for it later
        self.condition = condition
        self.if_empty = if_empty
        self.prepend = prepend
        self.append = append

    def __repr__(self):
        return "ConditionalValue(" + ", ".join(s for s in (
            "condition={}".format(self.condition) if self.condition else None,
            "if_empty=True" if self.if_empty else None,
            "prepend=True" if self.prepend else None,
            "append=True" if self.append else None,
        ) if s) + ")"

    def __str__(self):
        return str(self.value)


def Prepend(value):
    r"""Add a value that will prepend any existing value.

This is a helper function for ConditionalValue that sets `prepend` to True.
Remember to include the appropriate separator character in your value.
"""
    return ConditionalValue(value, prepend=True)


class Property:
    r"""Add a property to a project.

When generated into an MSBuild project, this property will be defined
at the point in member order where it is specified.

Use `ConditionalValue` to add conditions to the property.
"""
    members = ()

    def __init__(self, name, value):
        self.name = name
        self.value = value


class ItemDefinition:
    r"""Add an item definition to a project.

When generated into an MSBuild project, the item definition group will
be defined at the point in member order where it is specified.

Provide a list or tuple as the metadata value to include multiple
specifications.

Use `ConditionalValue` to add conditions to each metadata item.
"""
    members = ()

    def __init__(self, kind, **metadata):
        self.name = "ItemDefinition({})".format(kind)
        self.kind = kind
        self.options = metadata


class File:
    r"""Add a generic file to a package or project.

When added to `Package`, the file will be copied into the resulting
package directory.

When added to another project, behaviour will depend on how that
project treats "Content" elements.
"""
    _ITEMNAME = "Content"
    options = {}
    has_condition = False
    condition = None

    def __init__(self, source, name=None, **metadata):
        self.source = PurePath(source)
        self.name = name or self.source.name
        self.members = []
        self.options = {**self.options, **metadata}

    def excluding(self, pattern):
        self.has_condition = True
        self.exclude = pattern
        return self

    def if_(self, condition):
        self.has_condition = True
        self.condition = condition
        return self


class PyFile(File):
    r"""Add a Python source file to a package.

Currently does nothing special. One day will generate .pyc files.
"""
    options = {"GeneratePyc": True}


class Project(File):
    r"""Add a reference to an external MSBuild project.

This project should either provide Build, BuildSdist and BuildInPlace
targets, or import "$(PyMsbuildTargets)\common.targets". Otherwise,
files referenced or generated by this project are not automatically
included in sdists or wheels, or the projcet may cause the build to
fail.
"""
    _ITEMNAME = "Project"


class SourceFile(File):
    r"""Add a generic file to use for building.

These files will be included in the sdist, but will not be copied
in-place or included in wheels.
"""
    _ITEMNAME = "None"


class CSourceFile(File):
    r"""Add a C/C++ source file.

These files will be included in the sdist, but will not be copied
in-place or included in wheels, except as built output.

Incremental rebuilds will be triggered when these files are modified,
and each file is expected to produce a linkable file.
"""
    _ITEMNAME = "ClCompile"


class LinkFile(File):
    r"""Add a linker input file.

These files will be included in the sdist, but will not be copied
in-place or included in wheels, except as built output.

Incremental rebuilds will be triggered when these files are modified,
and each file will be linked into the final output.
"""
    _ITEMNAME = "Link"

class IncludeFile(File):
    r"""Add a header file.

These files will be included in the sdist, but will not be copied
in-place or included in wheels, except as built output.

Incremental rebuilds will be triggered when these files are modified,
but they do not produce linkable outputs.
"""
    _ITEMNAME = "ClInclude"

class RemoveFile:
    r"""Removes a file that has already been added.

This must appear after the files that were added, and must specify the
same kind (either the class used to create it, or the string literal
matching the MSBuild item name).
"""
    _ITEMNAME = None
    name = None
    members = ()
    has_condition = True
    condition = None

    def __init__(self, kind, pattern):
        self._ITEMNAME = getattr(kind, "_ITEMNAME", None) or str(kind)
        self.exclude = pattern

    def if_(self, condition):
        self.condition = condition

    def __str__(self):
        return ""



import os
import subprocess
import sys

from pathlib import Path


if __name__ == "__main__":
    sys.stdout.buffer.raw.write("\0".join(sys.argv[1:]).encode())
    sys.stdout.buffer.raw.flush()
    sys.exit(0)


def _split_args(argv):
    return subprocess.check_output(
        f'python "{__file__}" {argv}',
        encoding="utf-8",
        executable=sys.executable,
    ).split("\0")


def locate_msbuild():
    exe = os.getenv("MSBUILD", "")
    if exe:
        if Path(exe).is_file():
            return [exe]
        return _split_args(exe)
    
    for part in os.getenv("PATH", "").split(os.path.pathsep):
        p = Path(part)
        if p.is_dir():
            exe = p / "msbuild.exe"
            if exe.is_file():
                return [str(exe)]

    vswhere = Path(os.getenv("ProgramFiles(x86)"), "Microsoft Visual Studio", "Installer", "vswhere.exe")
    if vswhere.is_file():
        out = Path(subprocess.check_output([
            str(vswhere),
            "-nologo",
            "-property", "installationPath",
            "-latest",
            "-prerelease",
            "-products", "*",
            "-requires", "Microsoft.VisualStudio.Component.VC.Tools.x86.x64",
            "-utf8",
        ], encoding="utf-8", errors="strict").strip())
        if out.is_dir():
            exe = out / "MSBuild" / "Current" / "Bin" / "msbuild.exe"
            if exe.is_file():
                return [str(exe)]

    try:
        out = subprocess.check_output(
            ["dotnet", "build", "-version", "-nologo"],
            encoding="ascii",
            errors="replace",
        )
        version = int(out.partition(".")[0])
        if version >= 15:
            return ["dotnet", "build"]
    except Exception:
        pass

    # TODO: Also look for .NET Core SDK installation

    raise RuntimeError("Unable to locate msbuild.exe. Please provide it as %MSBUILD%")



import re
import sysconfig
import types

from packaging.tags import Tag, parse_tag, sys_tags


_TAG_PLATFORM_MAP = {
    "win32": "Win32",
    "win_amd64": "x64",
    "win_arm": "ARM",
    "win_arm64": "ARM64",
    "linux_x86_64": "POSIX_x64",
    "linux_aarch64": "POSIX_ARM64",
    "macosx_10_15_x86_64": "POSIX_x64",
    "any": None,
}


def remap_platform_to_msbuild(platform_tag):
    try:
        return _TAG_PLATFORM_MAP[platform_tag]
    except LookupError:
        if re.match(r"manylinux.+?_x86_64", platform_tag):
            return "POSIX_x64"
        raise


def remap_ext_to_abi(ext_tag):
    m = re.match(r"(cp\d+)-(win.+)", ext_tag)
    if m:
        # Matches Windows-style tag
        abi = "m" if m.group(1) in {"cp37"} else ""
        return "{0}{abi}-{1}".format(*m.groups(), abi=abi)
    m = re.match(r"cpython-(\d+m?)-(.+)-linux-gnu", ext_tag)
    if m:
        # Matches Linux/GNU style tag
        return "cp{0}-linux_{1}".format(*m.groups())
    return ext_tag


def remap_abi_to_ext(abi_tag):
    m = re.match(r"(cp\d+)m?-(win.+)", abi_tag)
    if m:
        # Matches Windows-style tag
        return "{0}-{1}".format(*m.groups())
    m = re.match(r"cp(\d+m?)-linux_(.+)", abi_tag)
    if m:
        # Matches Linux/GNU style tag
        return "cpython-{0}-{1}-linux-gnu".format(*m.groups())
    m = re.match(r"cp(\d+m?)-manylinux.+?_(.+)", abi_tag)
    if m:
        # Matches manylinux style tag
        return "cpython-{0}-{1}-linux-gnu".format(*m.groups())
    return abi_tag


def choose_best_tags(
    sys_wheel_tag=None,
    sys_ext_suffix=None,
    wheel_tag=None,
    abi_tag=None,
    platform_tag=None,
    ext_suffix=None
):
    if not sys_wheel_tag:
        sys_wheel_tag = next(iter(sys_tags()), None) or "py3-none-any"

    if not sys_ext_suffix:
        sys_ext_suffix = sysconfig.get_config_var("EXT_SUFFIX")
        if not sys_ext_suffix:
            import importlib.machinery
            sys_ext_suffix = importlib.machinery.EXTENSION_SUFFIXES[0]

    if sys_wheel_tag and not isinstance(sys_wheel_tag, Tag):
        sys_wheel_tag = next(iter(parse_tag(sys_wheel_tag)), None)

    if wheel_tag and not isinstance(wheel_tag, Tag):
        # Ensure we can correctly parse the tag, but then just split
        # We don't want to expand out compressed fields
        if next(iter(parse_tag(wheel_tag)), None):
            wheel_tag = Tag(*wheel_tag.split('-', 3))

    # Extract the ABI portion from an explicit ABI tag or wheel tag
    abi_only = None
    if abi_tag:
        abi_only = abi_tag.partition('-')[0]
    elif wheel_tag and wheel_tag.abi != "*" and "." not in wheel_tag.abi:
        abi_only = wheel_tag.abi

    # Overwrite the ABI tag and platform tag from an explicit wheel tag
    if wheel_tag:
        if not abi_tag and "*" not in (wheel_tag.abi, wheel_tag.platform):
            abi_tag = f"{wheel_tag.abi}-{wheel_tag.platform}"
            if "." in abi_tag:
                abi_tag = None
        if not platform_tag and wheel_tag.platform != "*" and "." not in wheel_tag.platform:
            platform_tag = wheel_tag.platform

    # Infer the ABI tag from explicit or default ext_suffix, and update the ABI
    # portion accordingly
    if not abi_tag:
        if ext_suffix:
            abi_tag = remap_ext_to_abi(ext_suffix.rpartition('.')[0].strip('.'))
        else:
            abi_tag = remap_ext_to_abi(sys_ext_suffix.rpartition('.')[0].strip('.'))
            if abi_tag and platform_tag:
                abi_tag = "{}-{}".format(abi_tag.partition('-')[0], platform_tag)
        if abi_tag:
            if abi_only:
                abi_tag = "{}-{}".format(abi_only, abi_tag.partition('-')[2])
            else:
                abi_only = abi_tag.partition('-')[0]

    # Infer the platform from the ABI tag or default wheel tag
    if not platform_tag:
        if abi_tag:
            platform_tag = abi_tag.partition('-')[2]
        else:
            platform_tag = sys_wheel_tag.platform

    # Infer the ext_suffix from the ABI tag or default ext_suffix
    if not ext_suffix:
        if abi_tag:
            ext_suffix = '.{}.{}'.format(
                remap_abi_to_ext(abi_tag),
                sys_ext_suffix.rpartition('.')[2]
            )
        else:
            ext_suffix = sys_ext_suffix

    # Infer the ABI portion from the default wheel tag
    if not abi_only:
        abi_only = sys_wheel_tag.abi

    # Final chance for ABI
    if not abi_tag:
        abi_tag = f"{abi_only}-{platform_tag}"

    if not abi_tag.startswith(abi_only):
        raise ValueError(f"Expected '{abi_tag}' to start with '{abi_only}'")

    # Infer any missing parts of the wheel tag
    if not wheel_tag:
        wheel_tag = Tag(sys_wheel_tag.interpreter, abi_only, platform_tag)
    if wheel_tag.interpreter == '*':
        wheel_tag = Tag(sys_wheel_tag.interpreter, wheel_tag.abi, wheel_tag.platform)
    if wheel_tag.abi == '*':
        wheel_tag = Tag(wheel_tag.interpreter, abi_only, wheel_tag.platform)
    if wheel_tag.platform == '*':
        wheel_tag = Tag(wheel_tag.interpreter, wheel_tag.abi, platform_tag)

    return types.SimpleNamespace(
        wheel_tag=str(wheel_tag),
        abi_tag=abi_tag,
        platform_tag=platform_tag,
        ext_suffix=ext_suffix,
    )



import py_compile
import os
import sys
from pathlib import Path, PureWindowsPath

try:
    PYC_OPTIMIZATION = int(os.getenv("PYMSBUILD_PYC_OPTIMIZE", "0"))
except ValueError:
    PYC_OPTIMIZATION = 0

RESID_COUNTER = iter(range(1001, 9999))
IMPORTERS_RESID = next(RESID_COUNTER)


def groupby(iterator, key):
    result = {}
    for i in iterator:
        result.setdefault(key(i), []).append(i)
    return result


def parse_all(file):
    g = groupby(map(str.strip, file), key=lambda i: i.partition(":")[0].lower())
    factories = dict(code=CodeFileInfo, resource=DataFileInfo, function=FunctionInfo)
    return [
        factories.get(k, ErrorInfo)(next(RESID_COUNTER), line)
        for k, v in g.items() for line in v
    ]


class CodeFileInfo:
    RC_TYPE = "PYCFILE"
    RC_TABLE = "IMPORT_TABLE"

    def __init__(self, resid, line):
        _, name, path = line.split(":", maxsplit=2)
        name = PureWindowsPath(name)
        path = Path(path)
        self.is_package = name.stem.casefold() == "__init__".casefold()
        self.name = ".".join(name.parts[:-1])
        if not self.is_package:
            self.name += "." + name.stem
        self.origin = str(name)
        self.sourcefile = path
        self._resource_file = None
        if path.suffix.casefold() == ".pyc".casefold():
            self._resource_file = self.sourcefile
        self.resid = resid

    def resource_file(self):
        if not self._resource_file:
            self._resource_file = Path(py_compile.compile(
                str(self.sourcefile),
                "pyc{}.bin".format(self.resid),
                self.origin,
                doraise=True,
                optimize=PYC_OPTIMIZATION,
                invalidation_mode=py_compile.PycInvalidationMode.UNCHECKED_HASH,
            ))
        return self._resource_file

    def check(self):
        if not self.sourcefile.is_file():
            return "Missing input: {}".format(self.sourcefile)

    @classmethod
    def get_builtin(cls, resid, sourcefile):
        return cls(resid, "code:${}:{}".format(sourcefile.stem, sourcefile))


class DataFileInfo:
    RC_TYPE = "DATAFILE"
    RC_TABLE = "DATA_TABLE"

    def __init__(self, resid, line):
        _, name, path = line.split(":", maxsplit=2)
        name = PureWindowsPath(name)
        path = Path(path)
        self.name = ".".join(name.parts)
        self.origin = str(name)
        self.sourcefile = path
        self.resid = resid

    def check(self):
        if not self.sourcefile.is_file():
            return "Missing input: {}".format(self.sourcefile)

    def resource_file(self):
        return self.sourcefile


class FunctionInfo:
    RC_TYPE = None
    RC_TABLE = "$FUNCTIONS"

    def __init__(self, resid, line):
        self.name = line.partition(":")[2]
        self.resid = resid

    def check(self):
        if not self.name.isidentifier():
            return "Invalid name: {}".format(self.name)

    def prototype(self):
        return "PyObject *{}(PyObject *, PyObject *, PyObject *);".format(self.name)


class ErrorInfo:
    RC_TYPE = None
    RC_TABLE = None

    def __init__(self, resid, line):
        self.line = line
        self.resid = resid

    def check(self):
        return "Unhandled input: " + line


def _write_rc_string(id, s, file):
    print(id, '"', end="", file=file)
    while len(s) > 4000:
        print(s[:4000], "\\", sep="", file=file)
        s = s[4000:]
    print(s, '"', sep="", file=file)


def _c_str(s):
    return '"' + str(s).replace("\\", "\\\\").replace('"', '\\"') + '"'


def _generate_files(module, files, targets):
    files.append(CodeFileInfo.get_builtin(IMPORTERS_RESID, targets / "dllpack_main.py"))

    with open("dllpack.rc", "w", encoding="ascii", errors="backslashescape") as rc_file:
        print("#define PYCFILE 257", file=rc_file)
        print("#define DATAFILE 258", file=rc_file)
        for f in files:
            if f.RC_TYPE:
                print(f.resid, f.RC_TYPE, _c_str(f.resource_file()), file=rc_file)

    with open("dllpack.h", "w", encoding="ascii", errors="backslashescape") as h_file:
        print('#define _MODULE_NAME "{}"'.format(module), file=h_file)
        print("#define _IMPORTERS_RESID", IMPORTERS_RESID, file=h_file)
        print('#define _INIT_FUNC_NAME PyInit_{}'.format(module), file=h_file)
        print("#define _PYCFILE 257", file=h_file)
        print("#define _DATAFILE 258", file=h_file)
        print("#define _PYC_HEADER_LEN 16", file=h_file)
        print("struct ENTRY {const char *name; const char *origin; int id;};", file=h_file)
        expected_tables = {"IMPORT_TABLE", "DATA_TABLE"}
        tables = groupby(files, lambda f: f.RC_TABLE)
        for table, table_files in tables.items():
            if not table or not table.isidentifier():
                continue
            expected_tables.discard(table)
            print("struct ENTRY ", table, "[] = {", sep="", file=h_file)
            for f in table_files:
                print("    {", file=h_file)
                print('        {},'.format(_c_str(f.name)), file=h_file)
                print('        {},'.format(_c_str(f.origin)), file=h_file)
                print("        {}".format(f.resid), file=h_file)
                print("    },", file=h_file)
            print("    {NULL, NULL, 0}", file=h_file)
            print("};", file=h_file)
        for table in expected_tables:
            print("struct ENTRY ", table, "[] = {{NULL, NULL, 0}};", sep="", file=h_file)
        for f in tables.get("$FUNCTIONS", ()):
            print("extern", f.prototype(), file=h_file);
        print("#define MOD_METH_TAIL \\", file=h_file)
        for f in tables.get("$FUNCTIONS", ()):
            print('    {{"{0}", (PyCFunction){0}, METH_VARARGS|METH_KEYWORDS, NULL}}, \\'.format(f.name), file=h_file)
        print("    {NULL, NULL, 0, NULL}", file=h_file)


if __name__ == "__main__":
    import sys
    MODULE = sys.argv[1]
    with open(sys.argv[2], "r", encoding="utf-8-sig") as f:
        PARSED = parse_all(f)
    ERRORS = [p.check() for p in PARSED]
    if any(ERRORS):
        print(*filter(None, ERRORS), sep="\n")
        sys.exit(1)
    TARGETS = Path(sys.argv[3]).absolute()
    _generate_files(MODULE, PARSED, TARGETS)



import os
import packaging.tags
import re
import shutil
import subprocess
import sys
import sysconfig

from pathlib import PurePath, Path

from . import _generate
from . import _tags

if sys.platform == "win32":
    _WINDOWS = True
    from ._locate_vs import locate_msbuild
else:
    _WINDOWS = False
    from ._locate_dotnet import locate_msbuild


# Needed to avoid printing an unhelpful message every time we invoke dotnet
os.environ["DOTNET_NOLOGO"] = "1"


def _add_and_record(zipfile, path, relpath, hashalg="sha256"):
    import base64, hashlib
    hasher = getattr(hashlib, hashalg)() if hashalg else None
    l = 0
    with open(path, "rb") as f:
        with zipfile.open(str(relpath), "w") as zf:
            for b in iter(lambda: f.read(8192), b""):
                if hasher:
                    hasher.update(b)
                l += len(b)
                zf.write(b)
    if hashalg:
        return "{},{}={},{}".format(
            relpath,
            hashalg,
            base64.urlsafe_b64encode(hasher.digest()).rstrip(b"=").decode(),
            l,
        )
    return "{},,".format(relpath)


def _quote(s, start='"', end='"'):
    if end and s.endswith("\\"):
        end = "\\" + end
    return start + s + end


class BuildState:
    # Updated around init_METADATA and init_PACKAGE calls
    current = None

    def __init__(self, output_dir=None):
        self._finalized = False
        self._is_sdist = False
        self.verbose = False
        self.quiet = False
        self.force = False
        self.config = None
        self.package = None
        self.metadata = None
        self.project = None
        self.configuration = None
        self.target = None
        self.msbuild_exe = None
        self.output_dir = Path(output_dir) if output_dir else None
        self.build_dir = None
        self.temp_dir = None
        self.layout_dir = None
        self.layout_extra_files = []
        self.pkginfo = None
        self.source_dir = Path.cwd()
        self.config_file = None
        self.targets = Path(__file__).absolute().parent / "targets"
        self.wheel_tag = None
        self.abi_tag = None
        self.ext_suffix = None
        self.platform = None
        self.build_number = None
        self.python_cflags = None
        self.python_ldflags = None
        self.python_includes = None
        self.python_libs = None

    def finalize(self, getenv=os.getenv):
        if self._finalized:
            return
        self._finalized = True

        self.output_dir = self.source_dir / (self.output_dir or "dist")
        self.build_dir = self.source_dir / (self.build_dir or "build/layout")
        self.temp_dir = self.source_dir / (self.temp_dir or "build/temp")
        self.pkginfo = self.source_dir / (self.pkginfo or "PKG-INFO")

        self._set_best("config_file", None, "PYMSBUILD_CONFIG", "_msbuild.py", getenv)

        type(self).current = self

        if self.config is None:
            import importlib.util
            file = self.source_dir / (self.config_file or "_msbuild.py")
            spec = importlib.util.spec_from_file_location("_msbuild", file)
            self.config = mod = importlib.util.module_from_spec(spec)
            mod.__loader__.exec_module(mod)

        if self.metadata is None:
            if self.pkginfo.is_file():
                self.log("Using", self.pkginfo)
                self.metadata = _generate.readback_distinfo(self.pkginfo)
                try:
                    self.config.METADATA.update(self.metadata)
                except AttributeError:
                    self.config.METADATA = self.metadata
            else:
                if hasattr(self.config, "init_METADATA"):
                    self.log("Dynamically initialising METADATA")
                    self.metadata = self.config.init_METADATA()
                    if self.metadata:
                        self.config.METADATA = self.metadata
                    else:
                        self.metadata = self.config.METADATA
                if hasattr(self.config, "METADATA"):
                    self.metadata = self.config.METADATA

        self._set_best("msbuild_exe", None, "MSBUILD", None, getenv)
        if self.msbuild_exe is None:
            self.msbuild_exe = locate_msbuild()
        if isinstance(self.msbuild_exe, str):
            if Path(self.msbuild_exe).is_file():
                self.msbuild_exe = [self.msbuild_exe]
            else:
                import shlex
                self.msbuild_exe = shlex.split(self.msbuild_exe)

        self._set_best("build_number", None, "BUILD_BUILDNUMBER", None, getenv)

        self._set_best("ext_suffix", "ExtSuffix", "PYMSBUILD_EXT_SUFFIX", None, getenv)
        self._set_best("abi_tag", "AbiTag", "PYMSBUILD_ABI_TAG", None, getenv)
        self._set_best("wheel_tag", "WheelTag", "PYMSBUILD_WHEEL_TAG", None, getenv)
        self._set_best("platform", None, "PYMSBUILD_PLATFORM", None, getenv)
        self._set_best("configuration", None, "PYMSBUILD_CONFIGURATION", "Release", getenv)

        tags = _tags.choose_best_tags(
            ext_suffix = self.ext_suffix,
            abi_tag = self.abi_tag,
            wheel_tag = self.wheel_tag,
            platform_tag = self.platform,
        )
        self.ext_suffix = tags.ext_suffix
        self.abi_tag = tags.abi_tag
        self.wheel_tag = tags.wheel_tag
        self.platform = tags.platform_tag

        self._set_best("python_config", None, "PYTHON_CONFIG", None, getenv)
        self._set_best("python_includes", None, "PYTHON_INCLUDES", getenv("PYMSBUILD_PYTHON_INCLUDES"), getenv)
        self._set_best("python_libs", None, "PYTHON_LIBS", getenv("PYMSBUILD_PYTHON_LIBS"), getenv)

        if not self.python_includes:
            self.python_includes = sysconfig.get_config_var("INCLUDEPY")
        if not self.python_libs:
            if _WINDOWS:
                self.python_libs = PurePath(sysconfig.get_config_var("installed_base")) / "libs"
            else:
                self.python_libs = sysconfig.get_config_var("LIBPL")

        if self.package is None:
            if hasattr(self.config, "init_PACKAGE"):
                self.log("Dynamically initialising PACKAGE")
                if self._is_sdist:
                    pack = self.config.init_PACKAGE(None)
                else:
                    pack = self.config.init_PACKAGE(str(self.wheel_tag))
                if pack:
                    self.config.PACKAGE = self.package
            self.package = self.config.PACKAGE

        type(self).current = None

    def _set_best(self, key, metakey, envkey, default, getenv):
        if getattr(self, key, None):
            self.log("Build state property", key, "already set to", getattr(self, key))
            return
        if metakey:
            v = self.metadata.get(metakey, None)
            if v:
                setattr(self, key, v)
                self.log("Build state property", key, "set to", v, "from metadata item", metakey)
                return
        if envkey:
            v = getenv(envkey)
            if v:
                setattr(self, key, v)
                self.log("Build state property", key, "set to", v, "from environment", envkey)
                return
        setattr(self, key, default)
        if default:
            self.log("Build state property", key, "set to default value", default)

    def log(self, *values, sep=" "):
        if self.verbose:
            print(*values, sep=sep)

    def write(self, *values, sep=" "):
        if not self.quiet:
            print(*values, sep=sep)

    def generate(self):
        if self.project:
            return self.project

        self.finalize()

        self.temp_dir.mkdir(parents=True, exist_ok=True)
        if self.metadata is not None:
            self.pkginfo = self.temp_dir / "PKG-INFO"
            self.log("Generating", self.pkginfo)
            _generate.generate_distinfo(self.metadata, self.temp_dir, self.source_dir)

        ext = self.ext_suffix
        self.log("Setting missing TargetExt to", ext)

        from . import _types as T
        for p in self.package:
            if isinstance(p, T.PydFile):
                p.options["TargetExt"] = p.options.get("TargetExt") or ext

        self.log("Generating projects")
        self.project = Path(_generate.generate(
            self.package,
            self.temp_dir,
            self.source_dir,
            self.config_file,
        ))
        self.log("Generated", self.project)

        return self.project

    def build(self, **properties):
        self.finalize()
        project = self.generate()
        if self.target is None:
            self.target = "Rebuild" if self.force else "Build"
        self.log("Compiling", project, "with", *self.msbuild_exe, "({})".format(self.target))
        if not project.is_file():
            raise FileNotFoundError(project)
        properties.setdefault("Configuration", self.configuration)
        if not properties.get("Platform"):
            try:
                properties["Platform"] = _tags.remap_platform_to_msbuild(self.platform)
            except LookupError:
                self.write("WARNING:", self.platform, "is not a known platform. Projects may not compile")
                properties["Platform"] = self.platform
        properties.setdefault("HostPython", sys.executable)
        properties.setdefault("PyMsbuildTargets", self.targets)
        properties.setdefault("_ProjectBuildTarget", self.target)
        properties.setdefault("SourceRootDir", self.source_dir)
        properties.setdefault("OutDir", self.build_dir)
        properties.setdefault("IntDir", self.temp_dir)
        properties.setdefault("PythonConfig", self.python_config)
        properties.setdefault("PythonIncludes", self.python_includes)
        properties.setdefault("PythonLibs", self.python_libs)
        rsp = self.temp_dir / f"{project}.{os.getpid()}.rsp"
        with rsp.open("w", encoding="utf-8-sig") as f:
            print(project, file=f)
            print("/nologo", file=f)
            if self.verbose:
                print("/p:_Low=Normal", file=f)
                print("/v:n", file=f)
            else:
                print("/v:m", file=f)
            print("/t:", self.target, sep="", file=f)
            for k, v in properties.items():
                if v is None:
                    continue
                if k in {"IntDir", "OutDir", "SourceDir"}:
                    v = f"{PurePath(v)}{os.path.sep}"
                print(_quote(f"/p:{k}={v}"), file=f)
        if self.verbose:
            with rsp.open("r", encoding="utf-8-sig") as f:
                self.log(" ".join(map(str.strip, f)))
            self.log()
        _run = subprocess.check_output if self.quiet else subprocess.check_call
        try:
            _run([*self.msbuild_exe, f"@{rsp}"], stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError as ex:
            if self.quiet:
                if _WINDOWS:
                    print(ex.stdout.decode("mbcs", "replace"))
                else:
                    print(ex.stdout.decode("utf-8", "replace"))
            sys.exit(1)
        else:
            try:
                rsp.unlink()
            except OSError:
                pass

    def build_in_place(self):
        self.finalize()
        self.generate()
        assert self.project
        self.target = "RebuildInPlace" if self.force else "BuildInPlace"
        self.build()

    def clean(self):
        self.finalize()
        p = self.config.PACKAGE
        self.project = self.temp_dir / (p.name + ".proj")
        if self.project.is_file():
            self.target = "Clean"
            self.build()

    def get_requires_for_build_sdist(self):
        self._is_sdist = True
        self.finalize()
        return self.metadata.get("BuildSdistRequires", [])

    def build_sdist(self):
        self._is_sdist = True
        self.finalize()
        self.target = "RebuildSdist" if self.force else "BuildSdist"
        if self.build_dir.is_dir():
            shutil.rmtree(self.build_dir)
            self.build_dir.mkdir(parents=True, exist_ok=True)
        self.generate()
        self.build()
        if self.layout_dir:
            return self.layout_sdist()
        return self.pack_sdist()

    def pack_sdist(self):
        self._is_sdist = True
        self.finalize()
        import gzip, tarfile
        self.output_dir.mkdir(parents=True, exist_ok=True)
        name, version = self.metadata["Name"], self.metadata["Version"]
        sdist_basename = "{}-{}".format(
            re.sub(r"[^\w\d.]+", "_", name, re.UNICODE),
            re.sub(r"[^\w\d.]+", "_", version, re.UNICODE),
        )
        sdist = self.output_dir / (sdist_basename + ".tar.gz")
        with gzip.open(sdist, "w") as f_gz:
            with tarfile.TarFile.open(
                sdist_basename + ".tar",
                "w",
                fileobj=f_gz,
                format=tarfile.PAX_FORMAT
            ) as f:
                f.add(
                    self.build_dir,
                    arcname=sdist_basename,
                    recursive=True,
                )
        self.write("Wrote sdist to", sdist)
        return sdist.name

    def get_requires_for_build_wheel(self):
        self._is_sdist = True
        self.finalize()
        return self.metadata.get("BuildWheelRequires", [])

    def build_wheel(self, metadata_dir=None):
        self.finalize()
        self.target = "Rebuild" if self.force else "Build"
        if self.build_dir.is_dir():
            shutil.rmtree(self.build_dir)
        if metadata_dir is None:
            metadata_dir = self.temp_dir / "metadata"
            if metadata_dir.is_dir():
                shutil.rmtree(metadata_dir)
        else:
            metadata_dir = Path(metadata_dir)

        self.generate()
        if not metadata_dir.is_dir():
            self.prepare_wheel_distinfo(metadata_dir)

        self.build()
        if self.layout_dir:
            return self.layout_wheel(metadata_dir)
        return self.pack_wheel(metadata_dir)

    def pack_wheel(self, metadata_dir):
        self.finalize()
        import zipfile
        self.output_dir.mkdir(parents=True, exist_ok=True)

        name, version = self.metadata["Name"], self.metadata["Version"]
        wheel = self.output_dir / "{}-{}-{}.whl".format(
            re.sub(r"[^\w\d.]+", "_", name, re.UNICODE),
            re.sub(r"[^\w\d.]+", "_", version, re.UNICODE),
            self.wheel_tag,
        )
        record = []
        with zipfile.ZipFile(wheel, "w", compression=zipfile.ZIP_DEFLATED) as f:
            for n in metadata_dir.rglob(r"**/*"):
                if n.is_file():
                    record.append(_add_and_record(f, n, n.relative_to(metadata_dir)))
            for n in self.build_dir.rglob(r"**/*"):
                if n.is_file():
                    record.append(_add_and_record(f, n, n.relative_to(self.build_dir)))
            record_files = []
            for n in metadata_dir.glob("*.dist-info"):
                if not n.is_dir():
                    continue
                record_files.append(r"{}/RECORD".format(n.name))
                record.append(r"{}/RECORD,,".format(n.name))
            record_file = "\n".join(record).encode("utf-8")
            for n in record_files:
                f.writestr(n, record_file)
        self.write("Wrote wheel to", wheel)
        return wheel.name

    def prepare_wheel_distinfo(self, metadata_dir=None):
        self.finalize()
        metadata_dir = Path(metadata_dir or self.output_dir)
        self.generate()
        name, version = self.metadata["Name"], self.metadata["Version"]
        outdir = metadata_dir / "{}-{}.dist-info".format(
            re.sub(r"[^\w\d.]+", "_", name, re.UNICODE),
            re.sub(r"[^\w\d.]+", "_", version, re.UNICODE),
        )
        outdir.mkdir(parents=True, exist_ok=True)
        from pymsbuild import __version__
        with open(outdir / "WHEEL", "w", encoding="utf-8") as f:
            print("Wheel-Version: 1.0", file=f)
            print("Generator: pymsbuild", __version__, file=f)
            print("Root-Is-Purelib: false", file=f)
            for t in sorted(self.wheel_tag):
                print("Tag:", t, file=f)
            if self.build_number:
                print("Build:", self.build_number, file=f)
        shutil.copy(self.pkginfo, outdir / "METADATA")
        return outdir.name

    def _write_state(self, state_file, outfile, outfmt):
        print("output-format=", outfmt, sep="", file=state_file)
        print("output-file=", outfile, sep="", file=state_file)
        for k in dir(self):
            if not k.startswith("_") and not hasattr(type(self), k):
                v = getattr(self, k)
                if isinstance(v, (str, PurePath)):
                    print(k, "=", getattr(self, k), sep="", file=state_file)
        print("# BEGIN FILES", file=state_file)

    def layout_sdist(self):
        self.finalize()
        import zipfile
        root = Path(self.layout_dir).absolute()
        assert root
        try:
            shutil.rmtree(root)
        except OSError:
            pass
        root.mkdir(parents=True, exist_ok=True)

        name, version = self.metadata["Name"], self.metadata["Version"]
        sdist_basename = "{}-{}".format(
            re.sub(r"[^\w\d.]+", "_", name, re.UNICODE),
            re.sub(r"[^\w\d.]+", "_", version, re.UNICODE),
        )
        sdist = self.output_dir / (sdist_basename + ".tar.gz")
        with (root / "__state.txt").open("w", encoding="utf-8") as f:
            self._write_state(f, sdist, "targz")
            for n in self.build_dir.rglob("**/*"):
                if n.is_file():
                    rn = root / n.relative_to(self.build_dir)
                    rn.parent.mkdir(parents=True, exist_ok=True)
                    shutil.copy(n, rn)
                    print(rn, file=f)
        self.write("Wrote layout to", root)
        return sdist.name

    def layout_wheel(self, metadata_dir):
        self.finalize()
        import zipfile
        root = Path(self.layout_dir).absolute()
        assert root
        try:
            shutil.rmtree(root)
        except OSError:
            pass
        root.mkdir(parents=True, exist_ok=True)

        name, version = self.metadata["Name"], self.metadata["Version"]
        wheel = self.output_dir / "{}-{}-{}.whl".format(
            re.sub(r"[^\w\d.]+", "_", name, re.UNICODE),
            re.sub(r"[^\w\d.]+", "_", version, re.UNICODE),
            self.wheel_tag,
        )
        with (root / "__state.txt").open("w", encoding="utf-8") as f:
            self._write_state(f, wheel, "whl")
            for n in metadata_dir.rglob("**/*"):
                if n.is_file():
                    rn = root / n.relative_to(metadata_dir)
                    rn.parent.mkdir(parents=True, exist_ok=True)
                    shutil.copy(n, rn)
                    print(rn, file=f)
            for n in self.build_dir.rglob("**/*"):
                if n.is_file():
                    rn = root / n.relative_to(self.build_dir)
                    rn.parent.mkdir(parents=True, exist_ok=True)
                    shutil.copy(n, rn)
                    print(rn, file=f)
        self.write("Wrote layout to", root)
        return wheel.name

    def pack(self):
        root = Path(self.layout_dir)
        if not root:
            print(
                "'--layout-dir' argument is required when invoking the 'pack' command",
                file=sys.stderr
            )
            return
        root = root.absolute()
        outfile = None
        outfmt = None
        with (root / "__state.txt").open("r", encoding="utf-8-sig") as f:
            for i in f:
                k, sep, v = i.strip().partition("=")
                if not sep:
                    break
                if k == "output-file":
                    outfile = Path(v)
                elif k == "output-format":
                    outfmt = v
                elif not k.startswith("_") and hasattr(self, k) and not hasattr(type(self), k):
                    setattr(self, k, v)
                else:
                    self.log("Property", k, "from layout directory is ignored")
            files = [root / i.strip() for i in f if i.strip()]
            extra = list(self.layout_extra_files or ())
            seen = set()
            while extra:
                i = (extra.pop(0) or "").strip()
                if i in seen:
                    continue
                seen.add(i)
                if i.startswith("@"):
                    with Path(i[1:]).open("r", encoding="utf-8-sig") as f2:
                        extra.extend(f2)
                elif i:
                    files.append(root / i)
        if not outfmt or not outfile:
            print("Layout appears to be corrupted. Please rerun the first stage again.", file=sys.stderr)
            return
        if outfmt == "targz":
            import gzip, tarfile
            outfile.parent.mkdir(parents=True, exist_ok=True)
            self.log("Writing .tar.gz to", outfile)
            with gzip.open(outfile, "w") as f_gz:
                with tarfile.TarFile.open(
                    PurePath(outfile).with_suffix(""),
                    "w",
                    fileobj=f_gz,
                    format=tarfile.PAX_FORMAT
                ) as f:
                    for n in files:
                        if n.is_file():
                            try:
                                rn = n.relative_to(root)
                            except ValueError:
                                self.write("Not including", n, "from outside of layout directory")
                            else:
                                self.log("Adding", rn)
                                f.add(n, arcname=rn)
            self.write("Wrote sdist to", outfile)
        elif outfmt == "whl":
            import zipfile
            record = []
            record_files = []
            outfile.parent.mkdir(parents=True, exist_ok=True)
            self.log("Writing .whl to", outfile)
            with zipfile.ZipFile(outfile, "w", compression=zipfile.ZIP_DEFLATED) as f:
                for n in files:
                    if n.is_file():
                        try:
                            rn = n.relative_to(root)
                        except ValueError:
                            self.write("Not including", n, "from outside of layout directory")
                        else:
                            self.log("Adding", rn)
                            record.append(_add_and_record(f, n, rn))
                    if n.match("*.dist-info/*"):
                        n2 = "{}{}RECORD".format(n.parent.name, os.path.sep)
                        if n2 not in record_files:
                            record_files.append(n2)
                            record.append("{},,".format(n2))
                record_file = "\n".join(record).encode("utf-8")
                for n in record_files:
                    f.writestr(n, record_file)
            self.write("Wrote wheel to", outfile)
        else:
            print(
                "Unsupported output format '{}'. Please do not modify the state file".format(
                    outfmt,
                ),
                file=sys.stderr,
            )



def _init():
    import sys

    from importlib import import_module
    from importlib.abc import Loader, MetaPathFinder, ResourceReader
    from importlib.machinery import ModuleSpec
    from ntpath import split as nt_split

    _NAME = __NAME()
    _NAME_DOT = _NAME + "."
    _MAKESPEC = __MAKESPEC
    _DATA = __DATA
    _DATA_NAMES = set(__DATA_NAMES())
    _CREATE_MODULE = __CREATE_MODULE
    _EXEC_MODULE = __EXEC_MODULE

    class DllPackReader(ResourceReader):
        def __init__(self, prefix):
            self.prefix = prefix

        def open_resource(self, resource):
            import io
            return io.BytesIO(_DATA(self.prefix + resource))

        def resource_path(self, resource):
            raise FileNotFoundError()

        def is_resource(self, resource):
            return prefix + resource in _DATA_NAMES

        def contents(self):
            p = self.prefix
            lp = len(p)
            return (n[lp:] for n in _DATA_NAMES if n.startswith(p))

    DllPackReader.__name__ += "_" + _NAME
    DllPackReader.__qualname__ = "<generated>." + DllPackReader.__name__

    class DllPackLoader(Loader):
        create_module = _CREATE_MODULE
        exec_module = _EXEC_MODULE

        def get_resource_reader(self, fullname):
            try:
                spec = _MAKESPEC(fullname, LOADER)
                if spec.submodule_search_locations is None:
                    return None
            except Exception:
                return None
            return DllPackReader(fullname + ".")

    DllPackLoader.__name__ += "_" + _NAME
    DllPackLoader.__qualname__ = "<generated>." + DllPackLoader.__name__

    LOADER = DllPackLoader()


    class DllPackFinder(MetaPathFinder):
        def find_spec(self, fullname, path, target=None):
            if fullname.startswith(_NAME_DOT) or fullname == _NAME:
                return _MAKESPEC(fullname, LOADER)

    DllPackFinder.__name__ += "_" + _NAME
    DllPackFinder.__qualname__ = "<generated>." + DllPackFinder.__name__

    FINDER = next((getattr(m, "__name__", None) == DllPackFinder.__name__ for m in sys.meta_path), None)
    if not FINDER:
        FINDER = DllPackFinder()
        sys.meta_path.append(FINDER)

    return _MAKESPEC(__name__, LOADER)


__spec__ = _init()
del _init, __NAME, __CREATE_MODULE, __EXEC_MODULE, __DATA, __DATA_NAMES, __MAKESPEC

__file__ = __spec__.origin
__loader__ = __spec__.loader
__path__ = []



from pathlib import Path
from pymsbuild._types import *

class CythonPydFile(PydFile):
    def __init__(self, name, *members, project_file=None, **kwargs):
        super().__init__(name, *members, project_file=project_file, **kwargs)
        self.members = [
            Property("BeforeBuildGenerateSourcesTargets", "Cythonize;$(BeforeBuildGenerateSourcesTargets)"),
            ItemDefinition("PyxCompile", TargetExt=".c", Dependencies=""),
            *self.members,
            LiteralXML('<Import Project="$(PyMsbuildTargets)/cython.targets" />'),
        ]


class PyxFile(File):
    _ITEMNAME = "PyxCompile"
    options = {"TargetExt": ".c"}


class CythonIncludeFile(File):
    _ITEMNAME = "CythonInclude"



import argparse
import pymsbuild
import sys

from os import getenv
from pathlib import PurePath, Path
from pymsbuild._build import BuildState


def _env(var, default=None):
    v = getenv(var)
    if not v:
        return default
    return v


def _envp(var, default=None):
    v = getenv(var)
    if not v:
        return default
    return PurePath(v)


def _envbool(var, default=None, is_true=True, is_false=False):
    v = getenv(var)
    if not v:
        return default
    if v.lower() in {"no", "0", "false"}:
        return is_false
    return is_true


def parse_args():
    parser = argparse.ArgumentParser("pymsbuild",)
    parser.add_argument(
        "--force", "-f", action="store_true", help="Force a full rebuild"
    )
    parser.add_argument(
        "--verbose", "-v", action="store_true", help="Additional output"
    )
    parser.add_argument("--quiet", "-q", action="store_true", help="Less output")
    parser.add_argument(
        "--debug", "-g", action="store_true", help="Build in debugging configuration"
    )
    parser.add_argument(
        "--source-dir",
        "-s",
        type=PurePath,
        default=_envp("PYMSBUILD_SOURCE_DIR"),
        help="Specify the source directory",
    )
    parser.add_argument(
        "--dist-dir",
        "-d",
        type=PurePath,
        default=_envp("PYMSBUILD_DIST_DIR"),
        help="Set the packaged outputs directory",
    )
    parser.add_argument(
        "--temp-dir",
        "-t",
        type=PurePath,
        default=_envp("PYMSBUILD_TEMP_DIR"),
        help="Set the temporary working directory",
    )
    parser.add_argument(
        "--layout-dir",
        type=PurePath,
        default=_envp("PYMSBUILD_LAYOUT_DIR"),
        help="Set the layout directory and enable two-step build",
    )
    parser.add_argument(
        "--config",
        "-c",
        type=PurePath,
        default=_envp("PYMSBUILD_CONFIG"),
        help="Override the path to _msbuild.py",
    )
    parser.add_argument(
        "--add",
        type=str,
        nargs="*",
        help="Specify additional file(s) to package when using the 'pack' command."
    )
    parser.add_argument(
        "command",
        type=str,
        nargs="*",
        help="""one or more of 'init', 'generate', 'sdist', 'wheel', 'pack', 'distinfo', 'clean'

init: Initialise a new _msbuild.py file.
generate: Generate the build files without building.
sdist: Build an sdist.
wheel: Build a wheel.
pack: Perform the second step of a two-step build.
distinfo: Build just the wheel metadata.
clean: Clean any builds.
""",
    )

    ns = parser.parse_args()
    if _envbool("PYMSBUILD_FORCE"):
        ns.force = True
    if _envbool("PYMSBUILD_VERBOSE"):
        ns.verbose = True
    if _envbool("PYMSBUILD_QUIET"):
        ns.quiet = True
    if _envbool("PYMSBUILD_DEBUG"):
        ns.debug = True

    return ns


ns = parse_args()
if not getattr(ns, "command", None):
    ns.command = ["build_in_place"]

if "init" in ns.command:
    from . import _init
    _init.run(Path.cwd(), ns.config)
    sys.exit(0)

bs = BuildState()
bs.source_dir = Path.cwd() / (ns.source_dir or "")
bs.output_dir = bs.source_dir / (ns.dist_dir or "dist")
bs.config_file = ns.config
root_dir = bs.source_dir / (ns.temp_dir or "build")
bs.build_dir = root_dir / "layout"
bs.temp_dir = root_dir / "temp"
bs.layout_dir = ns.layout_dir
bs.layout_extra_files = ns.add
bs.verbose = ns.verbose
bs.quiet = ns.quiet
bs.force = ns.force
if ns.debug:
    bs.configuration = "Debug"

for cmd in ns.command:
    cmd = {
        "sdist": "build_sdist",
        "wheel": "build_wheel",
        "distinfo": "prepare_wheel_distinfo",
    }.get(cmd, cmd)
    f = getattr(bs, cmd)
    f()



import os
import sys

from pathlib import PurePath, Path
from ._types import Package, PydFile, File, LiteralXML, Property, ItemDefinition, ConditionalValue
from ._writer import ProjectFileWriter

LIBPATH = Path(sys.base_prefix) / "libs"
INCPATH = Path(sys.base_prefix) / "include"
SEP = os.path.sep

class GroupSwitcher:
    def __init__(self, project):
        self.project = project
        self.tag = None
        self._cm = None

    def switch_to(self, tag):
        if tag == self.tag:
            return
        if self._cm:
            self._cm.__exit__(None, None, None)
            self._cm = None
        if tag:
            self._cm = self.project.group(tag)
            self._cm.__enter__()
        self.tag = tag

    def __enter__(self):
        return self

    def __exit__(self, *exc_info):
        if self._cm:
            return self._cm.__exit__(*exc_info)


def _resolve_wildcards(basename, source, include_dirs=True):
    if source.parent.name == "**":
        name = PurePath(basename).parent
        root = source.parent.parent
        for p in root.iterdir():
            if p.is_dir():
                yield from _resolve_wildcards(
                    name / p.relative_to(root) / source.name,
                    p / "**" / source.name,
                    include_dirs=False,
                )
            else:
                yield name / p.name, p
    elif "**" in source.parts:
        raise ValueError("Unsupported wildcard pattern" + str(source))
    elif "*" in source.name or "?" in source.name:
        name = PurePath(basename).parent
        for p in source.parent.glob(source.name):
            if not p.is_dir():
                yield name / p.name, p
    elif any("*" in p or "?" in p for p in source.parts):
        raise ValueError("Unsupported wildcard pattern " + str(source))
    elif include_dirs or not source.is_dir():
        yield basename, source


def _all_members(item, recurse_if=None, return_if=None, *, prefix=""):
    if not return_if or return_if(item):
        yield "{}{}".format(prefix, item.name), item
    if not recurse_if or recurse_if(item):
        for m in item.members:
            yield from _all_members(
                m,
                recurse_if,
                return_if,
                prefix="{}{}/".format(prefix, item.name),
            )


def _write_members(f, source_dir, members):
    with GroupSwitcher(f) as g:
        for n, p in members:
            if isinstance(p, File):
                g.switch_to("ItemGroup")
                wrote_any = False
                exclude = ()
                condition = None
                if getattr(p, "has_condition", False):
                    condition = getattr(p, "condition", None)
                    pattern = getattr(p, "exclude", None)
                    if pattern:
                        exclude = {p2 for n2, p2 in _resolve_wildcards(n, source_dir / pattern)}
                for n2, p2 in _resolve_wildcards(n, source_dir / p.source):
                    if p2 in exclude:
                        continue
                    options = dict(p.options)
                    options.setdefault("SourceDir", source_dir)
                    options.setdefault("Name", n2)
                    if condition:
                        p2 = ConditionalValue(p2, condition=condition)
                    f.add_item(p._ITEMNAME, p2, **options)
                    wrote_any = True
                if not wrote_any:
                    raise ValueError("failed to find any files for {} in {}".format(
                        p.source, source_dir))
            elif isinstance(p, Property):
                g.switch_to("PropertyGroup")
                f.add_property(p.name, p.value)
            elif isinstance(p, ItemDefinition):
                g.switch_to("ItemDefinitionGroup")
                with f.group(p.kind):
                    for k, v in p.options.items():
                        f.add_item_property(p.kind, k, v)
            elif isinstance(p, LiteralXML):
                g.switch_to(None)
                f.add_text(p.xml)
            elif hasattr(p, "write_member"):
                p.write_member(f, g)


def _generate_pyd(project, build_dir, root_dir):
    build_dir = Path(build_dir)
    proj = build_dir / "{}.proj".format(project.name)
    root_dir = Path(root_dir)
    source_dir = root_dir / project.source

    if project.project_file:
        return Path(project.project_file)

    tpath = project.options.get("TargetName", project.name) + (project.options.get("TargetExt") or ".pyd")
    tname, tdot, text = tpath.rpartition(".")
    with ProjectFileWriter(proj, tname, vc_platforms=True, root_namespace=project.name) as f:
        with f.group("PropertyGroup", Label="Globals"):
            f.add_property("SourceDir", ConditionalValue(source_dir, if_empty=True))
            f.add_property("SourceRootDir", ConditionalValue(root_dir, if_empty=True))
            f.add_property("OutDir", f"layout{SEP}")
            f.add_property("IntDir", ConditionalValue(f"build{SEP}", if_empty=True))
            f.add_property("__TargetExt", tdot + text)
            for k, v in project.options.items():
                if k not in {"ConfigurationType", "TargetExt"}:
                    f.add_property(k, v)
        f.add_import(f"$(PyMsbuildTargets){SEP}cpp-default-$(Platform).props")
        f.add_import(f"$(PyMsbuildTargets){SEP}common.props")
        with f.group("PropertyGroup", Label="Configuration"):
            f.add_property("ConfigurationType", project.options.get("ConfigurationType", "DynamicLibrary"))
            f.add_property("PlatformToolset", "$(DefaultPlatformToolset)")
            f.add_property("BasePlatformToolset", "$(DefaultPlatformToolset)")
            f.add_property("CharacterSet", "Unicode")
        f.add_import(f"$(PyMsbuildTargets){SEP}cpp-$(Platform).props")
        f.add_import(f"$(PyMsbuildTargets){SEP}pyd.props")

        _write_members(f, source_dir, _all_members(project, recurse_if=lambda m: m is project))
        for n, p in _all_members(project, recurse_if=lambda m: m is project, return_if=lambda m: isinstance(m, Package)):
            _write_members(
                f,
                source_dir,
                _all_members(p, recurse_if=lambda m: not isinstance(m, PydFile), prefix=f"{project.name}/")
            )

        f.add_import(f"$(PyMsbuildTargets){SEP}common.targets")
        f.add_import(f"$(PyMsbuildTargets){SEP}cpp-$(Platform).targets")
        f.add_import(f"$(PyMsbuildTargets){SEP}pyd.targets")

    return proj


def _generate_pyd_reference_metadata(relname, project, source_dir):
    output = project.options.get("TargetName", relname.stem) + project.options.get("TargetExt", ".pyd")
    tname, tdot, text = output.rpartition(".")
    return {
        **dict(
            TargetDir=relname.parent,
            IntDir="$(IntDir)" + tname,
        ),
        **project.options,
        **dict(
            TargetName=tname,
            TargetExt=tdot + text,
            SourceDir=source_dir / project.source,
        ),
    }


def generate(project, build_dir, source_dir, config_file=None):
    build_dir = Path(build_dir)
    root_dir = Path(source_dir)
    config_file = root_dir / (config_file or "_msbuild.py")
    source_dir = root_dir / project.source
    proj = build_dir / "{}.proj".format(project.name)

    if project.project_file:
        return Path(project.project_file)

    if isinstance(project, PydFile):
        return _generate_pyd(project, build_dir, root_dir)

    with ProjectFileWriter(proj, project.name) as f:
        with f.group("PropertyGroup"):
            f.add_property("SourceDir", ConditionalValue(source_dir, if_empty=True))
            f.add_property("SourceRootDir", ConditionalValue(root_dir, if_empty=True))
            f.add_property("OutDir", ConditionalValue(f"layout{SEP}", if_empty=True))
            f.add_property("IntDir", ConditionalValue(f"build{SEP}", if_empty=True))
            for k, v in project.options.items():
                f.add_property(k, v)
        f.add_import(f"$(PyMsbuildTargets){SEP}common.props")
        with f.group("ItemGroup", Label="ProjectReferences"):
            for n, p in _all_members(project, return_if=lambda m: m is not project and isinstance(m, PydFile)):
                fn = PurePath(n)
                pdir = _generate_pyd(p, build_dir, source_dir)
                try:
                    pdir = pdir.relative_to(proj.parent)
                except ValueError:
                    pass
                f.add_item(
                    "Project",
                    pdir,
                    Name=n,
                    **_generate_pyd_reference_metadata(fn, p, source_dir),
                )
        with f.group("ItemGroup", Label="Sdist metadata"):
            f.add_item("Sdist", build_dir / "PKG-INFO", RelativeSource="PKG-INFO")
            f.add_item("Sdist", config_file, RelativeSource="_msbuild.py")
            f.add_item("Sdist", root_dir / "pyproject.toml", RelativeSource="pyproject.toml")
        _write_members(f, source_dir, _all_members(
            project,
            recurse_if=lambda m: not isinstance(m, PydFile),
        ))
        f.add_import(f"$(PyMsbuildTargets){SEP}common.targets")
        f.add_import(f"$(PyMsbuildTargets){SEP}package.targets")

    return proj


def _write_metadata(f, key, value, source_dir):
    if not isinstance(value, str) and hasattr(value, "__iter__"):
        for v in value:
            _write_metadata(f, key, v, source_dir)
        return
    if isinstance(value, File):
        value = (source_dir / value.source).read_text(encoding="utf-8")
    if "\n" in value:
        value = value.replace("\n", "\n       |")
    print(key, value, sep=": ", file=f)


def _write_metadata_description(f, value, source_dir):
    if not isinstance(value, str) and hasattr(value, "__iter__"):
        for v in value:
            _write_metadata_description(f, v, source_dir)
        return
    if isinstance(value, File):
        value = (source_dir / value.source).read_text(encoding="utf-8")
    print(file=f)
    print(value, file=f)


def generate_distinfo(distinfo, build_dir, source_dir):
    build_dir.mkdir(parents=True, exist_ok=True)
    with (build_dir / "PKG-INFO").open("w", encoding="utf-8") as f:
        description = None
        for k, vv in distinfo.items():
            if k.casefold() == "description".casefold():
                description = vv
                continue
            _write_metadata(f, k, vv, source_dir)
        if description:
            _write_metadata_description(f, description, source_dir)

def readback_distinfo(pkg_info):
    distinfo = []
    description = None
    with open(pkg_info, "r", encoding="utf-8") as f:
        for line in f:
            if description is not None:
                description.append(line)
                continue
            line = line.rstrip()
            if not line:
                description = []
                continue
            if line.startswith(("       |", "        ")):
                distinfo[-1] = (distinfo[-1][0], distinfo[-1][1] + "\n" + line[8:])
                continue
            key, _, value = line.partition(":")
            distinfo.append((key.strip(), value.lstrip()))
    d = {}
    for k, v in distinfo:
        r = d.setdefault(k, v)
        if isinstance(r, list):
            r.append(v)
        elif r is not v:
            d[k] = [r, v]
    if description:
        d["Description"] = "".join(description).rstrip()
    return d



import os
from pymsbuild import *

METADATA = {
    "Metadata-Version": "2.1",
    "Name": "pymsbuild",
    "Version": "0.0.1",
    "Author": "Steve Dower",
    "Author-email": "steve.dower@python.org",
    "Home-page": "https://github.com/zooba/pymsbuild",
    "Project-url": [
        "Bug Tracker, https://github.com/zooba/pymsbuild/issues",
    ],
    "Summary": "The pymsbuild build backend.",
    "Description": File("README.md"),
    "Description-Content-Type": "text/markdown",
    "Keywords": "build,pep-517,msbuild,packaging,cython",
    "Classifier": [
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Topic :: Software Development :: Build Tools",
        "Topic :: Software Development :: Compilers",
        "Topic :: Utilities",
    ],
    "Requires-Python": ">=3.7",
    "Requires-External": "msbuild,dotnet",
    "Requires-Dist": [
        "packaging",
    ],
    "WheelTag": "py3-none-any",
}

PACKAGE = Package(
    "pymsbuild",
    PyFile("pymsbuild/*.py"),
    File("pymsbuild/targets/*", name="targets/*"),
)

def init_METADATA():
    version = os.getenv("BUILD_BUILDNUMBER")
    ghref = os.getenv("GITHUB_REF")
    if ghref:
        version = ghref.rpartition("/")[2]
    if version:
        METADATA["Version"] = version

def init_PACKAGE(tag=None):
    if tag is None:
        return
    with open("pymsbuild/__init__.py", "r", encoding="utf-8") as f:
        content = f.read()
    content = content.replace("%VERSION%", METADATA["Version"])
    with open("pymsbuild/__init__.py.ver", "w", encoding="utf-8") as f:
        f.write(content)
    PACKAGE.members.append(RemoveFile(PyFile, "pymsbuild/__init__.py"))
    PACKAGE.members.append(PyFile("pymsbuild/__init__.py.ver", name="__init__.py"))



import contextlib
import uuid

from pathlib import Path

_GENERATED_NAMESPACE = uuid.UUID('db509c23-800c-41d5-9d00-359fc120e87a')

PROLOGUE = r"""<?xml version="1.0" encoding="utf-8"?>
<Project DefaultTargets="Build" ToolsVersion="Current" xmlns="http://schemas.microsoft.com/developer/msbuild/2003">"""


TARGETS = Path(__file__).parent / "targets"


def _guid(target_name):
    return uuid.uuid3(_GENERATED_NAMESPACE, target_name)


class CV:
    has_condition = True

    def __init__(self, value, condition=None, if_empty=False):
        self.value = str(value)
        self.condition = condition
        self.if_empty = if_empty

    def __str__(self):
        return self.value


class ProjectFileWriter:
    def __init__(self, filename, target_name, *, vc_platforms=None, root_namespace=None):
        self.filename = filename
        self.target_name = target_name
        self.root_namespace = root_namespace or target_name
        self._file = None
        self._vc_platforms = vc_platforms
        self.indent = 2
        self.current_group = None

    def __enter__(self):
        Path(self.filename).parent.mkdir(parents=True, exist_ok=True)
        self._file = open(self.filename, "w", encoding="utf-8")
        print(PROLOGUE, file=self._file)
        if self._vc_platforms is True:
            self.add_vc_platforms()
        elif self._vc_platforms:
            self.add_vc_platforms(*self._vc_platforms)
        with self.group("PropertyGroup", Label="Globals"):
            self.add_property("Configuration", CV("Release", "$(Configuration) == ''"))
            self.add_property("Platform", CV("x64", "$(Platform) == ''"))
            self.add_property("ProjectGuid", _guid(self.target_name))
            self.add_property("RootNamespace", self.root_namespace)
            self.add_property("TargetName", self.target_name)
            self.add_property("PyMsbuildTargets", CV(TARGETS, if_empty=True))
        return self

    def __exit__(self, *exc_info):
        print("</Project>", file=self._file)
        self._file.flush()
        self._file.close()
        self._file = None

    def write(self, *text):
        print(" " * self.indent, *text, sep="", file=self._file)

    @contextlib.contextmanager
    def group(self, tag, **attributes):
        if attributes:
            self.write("<", tag, *(' {}="{}"'.format(*i) for i in attributes.items() if all(i)), ">")
        else:
            self.write("<", tag, ">")
        self.indent += 2
        old_group, self.current_group = self.current_group, tag
        yield
        self.current_group = old_group
        self.indent -= 2
        self.write("</", tag, ">")

    def _write_value(self, name, value, symbol='$'):
        if isinstance(value, (tuple, list)):
            for v in value:
                self._write_value(name, v, symbol)
            return
        c = None
        v = str(value)
        if getattr(value, "has_condition", None):
            c = value.condition
            if getattr(value, "if_empty", False):
                c = "{}({}) == ''".format(symbol, name)
            if getattr(value, "append", False):
                v = "{}({}){}".format(symbol, name, v)
            if getattr(value, "prepend", False):
                v = "{}{}({})".format(v, symbol, name)
        if c:
            self.write("<", name, ' Condition="', c, '">', v, "</", name, ">")
        else:
            self.write("<", name, ">", v, "</", name, ">")

    def add_property(self, name, value):
        self._write_value(name, value, "$")

    def add_item(self, kind, name, **metadata):
        c = None
        excl = None
        remove = None
        if getattr(name, "has_condition", False):
            c = name.condition
            excl = getattr(name, "exclude", None)
            remove = getattr(name, "remove", None)
            if getattr(name, "if_empty", False):
                c = "@({}) == ''".format(kind)
            if getattr(name, "append", False) or getattr(name, "prepend", False):
                raise ValueError("'append' and 'prepend' are not supported on '{}'".format(name))

        attrs = dict(Include=str(name), Condition=c, Exclude=excl, Remove=remove)
        attrs = {k: v for k, v in attrs.items() if v}
        if metadata:
            with self.group(kind, **attrs):
                for k, v in metadata.items():
                    if v is not None:
                        self._write_value(k, v, "%")
        else:
            self.write("<", kind, *[f' {k}="{v}"' for k, v in attrs.items()], " />")

    def add_item_property(self, kind, name, value):
        self._write_value(name, value, "%")

    def add_import(self, project):
        self.write('<Import Project="', project, '" />')

    def add_vc_platforms(self, platforms=None, configurations=None):
        if not platforms:
            platforms = ["Win32", "x64", "ARM", "ARM64"]
        if not configurations:
            configurations = ["Debug", "Release"]
        with self.group("ItemGroup", Label="ProjectConfigurations"):
            for c in configurations:
                for p in platforms:
                    with self.group("ProjectConfiguration", Include="{}|{}".format(c, p)):
                        self.add_property("Configuration", c)
                        self.add_property("Platform", p)

    def add_text(self, text):
        for line in text.splitlines():
            self.write(line)



from pathlib import Path

def read_sysconfig_from_file(text):
    import ast

    data = {}

    class Visitor(ast.NodeVisitor):
        def visit_Assign(self, n):
            if (n.targets
                and isinstance(n.targets[0], ast.Name)
                and n.targets[0].id == "build_time_vars"
                and isinstance(n.value, ast.Dict)
            ):
                for k, v in zip(n.value.keys, n.value.values):
                    if isinstance(k, ast.Constant) and isinstance(v, ast.Constant):
                        data[k.value] = v.value
                    # Handle pre-3.8, just in case
                    elif hasattr(ast, "Str"):
                        if isinstance(k, ast.Str) and isinstance(v, ast.Str):
                            data[k.s] = v.s

    tree = ast.parse(text)
    Visitor().visit(tree)
    return data


def load_sysconfig(file=None):
    if file:
        data = read_sysconfig_from_file(Path(file).read_text(encoding="utf-8"))
        data["__FROM_FILE"] = str(file)
    else:
        import sysconfig
        data = sysconfig.get_config_vars()
        for k in [k for k in data if k.startswith(("HAVE_", "DOUBLE_IS_", "PY_SSL"))]:
            data.pop(k, None)

    if "SO" not in data:
        data["SO"] = f".{data['SOABI']}{data['SHLIB_SUFFIX']}"

    return data


if __name__ == "__main__":
    import pprint
    import sys
    pprint.pprint(load_sysconfig(sys.argv[1] if len(sys.argv) > 1 else None))



from pymsbuild._types import *

class DllPackage(PydFile):
    r"""Represents a DLL-packed package.

This is the equivalent of a regular `Package`, but the output is a
compiled DLL that exposes submodules and resources using an import hook.

Add `Function` elements to link """
    options = {
        **PydFile.options,
    }

    def __init__(self, name, *members, project_file=None, **kwargs):
        super().__init__(
            name,
            *members,
            LiteralXML('<Import Project="$(PyMsbuildTargets)\\dllpack.targets" />'),
            project_file=project_file,
            **kwargs
        )


class CFunction:
    r"""Represents a function exposed in a DLL-packed package.

The named function must be provided in a `CSourceFile` element and
follow this prototype:

```
PyObject *function(PyObject *module, PyObject *args, PyObject *kwargs)
```

It will be available in the root of the package as the same name.
"""
    _ITEMNAME = "DllPackFunction"

    def __init__(self, name, **options):
        self.name = name
        self.options = dict(**options)

    def write_member(self, project, group):
        group.switch_to("ItemGroup")
        project.add_item(self._ITEMNAME, self.name, **self.options)



"""The pymsbuild build backend.
"""

__version__ = "%VERSION%"

PYMSBUILD_REQUIRES_SPEC = f"pymsbuild>={__version__},<1.0"

from pymsbuild._build import BuildState as _BuildState
from pymsbuild._types import *


def get_current_build_state():
    return _BuildState.current


def build_sdist(sdist_directory, config_settings=None):
    bs = _BuildState(sdist_directory)
    return bs.build_sdist()


def build_wheel(wheel_directory, config_settings=None, metadata_directory=None):
    bs = _BuildState(wheel_directory)
    return bs.build_wheel()


def prepare_metadata_for_build_wheel(metadata_directory, config_settings=None):
    bs = _BuildState(metadata_directory)
    return bs.prepare_wheel_distinfo(bs.output_dir)


def get_requires_for_build_sdist(config_settings=None):
    bs = _BuildState()
    return bs.get_requires_for_build_sdist()


def get_requires_for_build_wheel(config_settings=None):
    bs = _BuildState()
    return bs.get_requires_for_build_wheel()



import importlib.resources
import itertools
import re
import sys

from . import PYMSBUILD_REQUIRES_SPEC

TEMPLATE = importlib.resources.read_text("pymsbuild", "_msbuild.py.in")
TOML_TEMPLATE = importlib.resources.read_text("pymsbuild", "pyproject.toml.in")

C_PREPROC_BLURB = """
# Need to set preprocessor variables or include dirs? Use a ClCompile item definition
#ItemDefinition(
#    "ClCompile",
#    PreprocessorDefinitions=Prepend("NAME=1;"),
#    AdditionalIncludeDirectories=Prepend("PATH;"),
#),
# Need to set linker options or directories? Use a Link item definition
#ItemDefinition(
#    "Link",
#    AdditionalDependencies=Prepend("kernel32.lib;"),
#    AdditionalLibraryDirectories=Prepend("PATH;"),
#),
""".strip().splitlines()

def _generate_module(root, offset=None, build_requires=None, _indent="    ", _root=None):
    if not root:
        return
    if not _root:
        _root = root
    yield f'{_indent}{root.name!r},'
    yield f'{_indent}PyFile("*.py"),'
    any_pyi = False
    for d in root.iterdir():
        if (d / "__init__.py").is_file():
            yield f''
            yield f'{_indent}# Discovered from {d.relative_to(_root)}'
            yield f'{_indent}Package('
            yield from _generate_module(d, build_requires=build_requires, _indent=_indent + '    ', _root=_root)
            yield f'{_indent}),'
        elif d.is_file():
            if d.match("*.pyx"):
                if "Cython" not in build_requires:
                    build_requires.append("Cython")
                yield f''
                yield f'{_indent}# Discovered from {d.relative_to(_root)}'
                yield f'{_indent}CythonPydFile('
                yield f'{_indent}    {d.stem!r},'
                yield from (f'{_indent}    {i}' for i in C_PREPROC_BLURB)
                yield f'{_indent}    PyxFile({d.name!r}),'
                if any(root.glob("*.pxi")):
                    yield f'{_indent}    CythonHeaderFile("*.pxi"),'
                yield f'{_indent}),'
            elif d.match("*.c") or d.match("*.cpp"):
                yield f''
                yield f'{_indent}# Discovered from {d.relative_to(_root)}'
                yield f'{_indent}PydFile('
                yield f'{_indent}    {d.stem!r},'
                yield from (f'{_indent}    {i}' for i in C_PREPROC_BLURB)
                yield f'{_indent}    CSourceFile({d.name!r}),'
                if any(root.glob("*.h")):
                    yield f'{_indent}    HeaderFile("*.h"),'
                if any(root.glob("*.hpp")):
                    yield f'{_indent}    HeaderFile("*.hpp"),'
                yield f'{_indent}),'
            elif d.match("*.pyi"):
                any_pyi = True
    if any_pyi:
        yield f'{_indent}File("*.pyi"),'
    if offset:
        yield f'{_indent}source={offset!r},'


def run(root, config_name="_msbuild.py"):
    if (root / config_name).is_file():
        print(config_name, "already exists. Delete the file before using 'init'", file=sys.stderr)
        return

    substitutions = {}
    build_requires = [PYMSBUILD_REQUIRES_SPEC]

    project = root
    offset = None
    if (root / "src").is_dir():
        offset = "src"
        project = root / "src"
    modules = [d for d in project.iterdir() if (d / "__main__.py").is_file()]
    if not modules:
        modules = [d for d in project.iterdir() if (d / "__init__.py").is_file()]
    if len(modules) == 1:
        project = modules[0]
    elif len(modules) > 1:
        project = max(modules, key=lambda m: len(list(m.glob("*.py"))))
    else:
        project = None
    substitutions["NAME"] = project.name if project else "package"

    substitutions["PACKAGE"] = "\n".join([
        "PACKAGE = Package(",
        *_generate_module(project, offset, build_requires, _root=project.parent),
        ")",
    ])

    substitutions["BUILD_REQUIRES"] = repr(build_requires)

    code = re.sub(r"\<(\w+)\>", lambda m: substitutions.get(m.group(1)), TEMPLATE)
    with open(root / config_name, "w", encoding="utf-8") as f:
        print(code, file=f, end="")
    print("Wrote", root / config_name)

    toml = re.sub(r"\<(\w+)\>", lambda m: substitutions.get(m.group(1)), TOML_TEMPLATE)
    pyproject = (root / config_name).parent / "pyproject.toml"
    if pyproject.is_file():
        pyproject = pyproject.parent / "pyproject.toml.txt"
        count = 1
        while pyproject.is_file():
            pyproject = pyproject.parent / f"pyproject.toml.{count}.txt"
            count += 1
        print("NOTE: pyproject.toml exists, so wrote recommended settings to", pyproject, file=sys.stderr)
    with open(pyproject, "w", encoding="utf-8") as f:
        print(toml, file=f, end="")
    print("Wrote", pyproject)



import os
import shlex
import shutil
import subprocess

from pathlib import Path
from urllib.request import urlretrieve

def _check_build(dotnet):
    try:
        out = subprocess.check_output(
            [dotnet, "build", "-version", "-nologo"],
            encoding="ascii",
            errors="replace",
        )
        version = int(out.partition(".")[0])
        if version >= 15:
            return [dotnet, "build"]
    except Exception:
        raise RuntimeError("Unable to locate 'dotnet build'. Please provide it as %MSBUILD%")

    raise RuntimeError("Unable to locate 'dotnet build'. Please provide it as %MSBUILD%")


def locate_msbuild():
    exe = os.getenv("MSBUILD", "")
    if exe:
        return shlex.split(exe)

    return _check_build("dotnet")
